@@ -25,9 +25,10 @@
 #endif
 
 #include <sys/types.h>
+#include <stdio.h>
 #include <errno.h>
 #include <limits.h>
-#include <stdio.h>
+#include <alloca.h>
 
 #include <ell/ell.h>
 
@@ -40,15 +41,16 @@
 #include "src/scan.h"
 #include "src/dbus.h"
 #include "src/agent.h"
-#include "src/device.h"
+#include "src/netdev.h"
 #include "src/wiphy.h"
+#include "src/station.h"
 #include "src/eap.h"
 #include "src/knownnetworks.h"
 #include "src/network.h"
 
 struct network {
 	char *object_path;
-	struct device *device;
+	struct station *station;
 	struct network_info *info;
 	unsigned char *psk;
 	char *passphrase;
@@ -57,7 +59,7 @@ struct network {
 	struct l_settings *settings;
 	struct l_queue *secrets;
 	bool update_psk:1;  /* Whether PSK should be written to storage */
-	bool ask_psk:1; /* Whether we should force-ask agent for PSK */
+	bool ask_passphrase:1; /* Whether we should force-ask agent */
 	int rank;
 };
 
@@ -107,7 +109,12 @@ static bool network_secret_check_cacheable(void *data, void *user_data)
 {
 	struct eap_secret_info *secret = data;
 
-	return secret->cache_policy == EAP_CACHE_NEVER;
+	if (secret->cache_policy == EAP_CACHE_NEVER) {
+		eap_secret_info_free(secret);
+		return true;
+	}
+
+	return false;
 }
 
 void network_connected(struct network *network)
@@ -127,11 +134,12 @@ void network_connected(struct network *network)
 		/*
 		 * This is an open network seen for the first time:
 		 *
-		 * Write an empty settings file to keep track of the
+		 * Write a settings file to keep track of the
 		 * last connected time.  This will also make iwd autoconnect
 		 * to this network in the future.
 		 */
-		network->settings = l_settings_new();
+		if (!network->settings)
+			network->settings = l_settings_new();
 
 		storage_network_sync(network_get_security(network),
 					network->info->ssid,
@@ -275,13 +283,13 @@ static void network_info_put(struct network_info *network)
 	network_info_free(network);
 }
 
-struct network *network_create(struct device *device, const char *ssid,
+struct network *network_create(struct station *station, const char *ssid,
 				enum security security)
 {
 	struct network *network;
 
 	network = l_new(struct network, 1);
-	network->device = device;
+	network->station = station;
 	network->info = network_info_get(ssid, security);
 
 	network->bss_list = l_queue_new();
@@ -294,11 +302,6 @@ const char *network_get_ssid(const struct network *network)
 	return network->info->ssid;
 }
 
-struct device *network_get_device(const struct network *network)
-{
-	return network->device;
-}
-
 const char *network_get_path(const struct network *network)
 {
 	return network->object_path;
@@ -309,7 +312,7 @@ enum security network_get_security(const struct network *network)
 	return network->info->type;
 }
 
-const uint8_t *network_get_psk(const struct network *network)
+const uint8_t *network_get_psk(struct network *network)
 {
 	return network->psk;
 }
@@ -396,36 +399,66 @@ static bool network_set_8021x_secrets(struct network *network)
 	return true;
 }
 
-static int network_load_psk(struct network *network)
+static int network_load_psk(struct network *network, bool need_passphrase)
 {
 	size_t len;
 	const char *psk = l_settings_get_value(network->settings,
 						"Security", "PreSharedKey");
+	char *passphrase = l_settings_get_string(network->settings,
+						"Security", "Passphrase");
+	struct network_info *info = network->info;
+	int r;
 
-	if (!psk)
+	/* PSK can be generated from the passphrase but not the other way */
+	if ((!psk || need_passphrase) && !passphrase)
 		return -ENOKEY;
 
+	l_free(network->passphrase);
+	network->passphrase = passphrase;
 	l_free(network->psk);
-	network->psk = l_util_from_hexstring(psk, &len);
-	if (!network->psk)
-		return -EINVAL;
 
-	if (len != 32) {
+	if (psk) {
+		char *path;
+
+		network->psk = l_util_from_hexstring(psk, &len);
+		if (network->psk && len == 32)
+			return 0;
+
+		path = storage_get_network_file_path(info->type, info->ssid);
+		l_error("%s: invalid PreSharedKey format", path);
+		l_free(path);
+
+		if (!passphrase)
+			goto reset_psk;
+
 		l_free(network->psk);
-		network->psk = NULL;
-		return -EINVAL;
 	}
 
-	l_free(network->passphrase);
-	network->passphrase = l_settings_get_string(network->settings,
-						"Security", "Passphrase");
+	network->psk = l_malloc(32);
+	r = crypto_psk_from_passphrase(passphrase, (uint8_t *) info->ssid,
+					strlen(info->ssid), network->psk);
+	if (!r) {
+		network->update_psk = true;
+		return 0;
+	}
 
-	return 0;
+	if (r == -ERANGE || r == -EINVAL)
+		l_error("PSK generation failed: invalid passphrase format");
+	else
+		l_error("PSK generation failed: %s.  "
+			"Ensure Crypto Engine is properly configured",
+			strerror(-r));
+
+	l_free(network->passphrase);
+	network->passphrase = NULL;
+reset_psk:
+	l_free(network->psk);
+	network->psk = NULL;
+	return -EINVAL;
 }
 
 void network_sync_psk(struct network *network)
 {
-	char *hex;
 	struct l_settings *fs_settings;
 
 	if (!network->update_psk)
@@ -433,47 +466,75 @@ void network_sync_psk(struct network *network)
 
 	network->update_psk = false;
 
-	hex = l_util_hexstring(network->psk, 32);
-	l_settings_set_value(network->settings, "Security",
+	fs_settings = storage_network_open(SECURITY_PSK, network->info->ssid);
+
+	if (network->psk) {
+		char *hex = l_util_hexstring(network->psk, 32);
+		l_settings_set_value(network->settings, "Security",
 						"PreSharedKey", hex);
 
-	if (network->passphrase)
+		if (fs_settings)
+			l_settings_set_value(fs_settings, "Security",
+						"PreSharedKey", hex);
+
+		l_free(hex);
+	}
+
+	if (network->passphrase) {
 		l_settings_set_string(network->settings, "Security",
 							"Passphrase",
 							network->passphrase);
 
-	fs_settings = storage_network_open(SECURITY_PSK, network->info->ssid);
-	if (fs_settings) {
-		l_settings_set_value(fs_settings, "Security",
-						"PreSharedKey", hex);
-		if (network->passphrase)
+		if (fs_settings)
 			l_settings_set_string(fs_settings, "Security",
 							"Passphrase",
 							network->passphrase);
+	}
 
+	if (fs_settings) {
 		storage_network_sync(SECURITY_PSK, network->info->ssid,
 					fs_settings);
 		l_settings_free(fs_settings);
 	} else
 		storage_network_sync(SECURITY_PSK, network->info->ssid,
 					network->settings);
+}
 
-	l_free(hex);
+static inline bool __bss_is_sae(const struct scan_bss *bss,
+						const struct ie_rsn_info *rsn)
+{
+	if (rsn->akm_suites & IE_RSN_AKM_SUITE_SAE_SHA256)
+		return true;
+
+	return false;
+}
+
+static bool bss_is_sae(const struct scan_bss *bss)
+{
+	struct ie_rsn_info rsn;
+
+	memset(&rsn, 0, sizeof(rsn));
+	scan_bss_get_rsn_info(bss, &rsn);
+
+	return __bss_is_sae(bss, &rsn);
 }
 
 int network_autoconnect(struct network *network, struct scan_bss *bss)
 {
-	struct wiphy *wiphy = device_get_wiphy(network->device);
+	struct station *station = network->station;
+	struct wiphy *wiphy = station_get_wiphy(station);
+	enum security security = network_get_security(network);
+	struct ie_rsn_info rsn;
 	bool is_autoconnectable;
 	bool is_rsn;
 	int ret;
 
-	switch (network_get_security(network)) {
+	switch (security) {
 	case SECURITY_NONE:
 		is_rsn = false;
 		break;
 	case SECURITY_PSK:
-		if (network->ask_psk)
+		if (network->ask_passphrase)
 			return -ENOKEY;
 
 		/* Fall through */
@@ -487,27 +548,6 @@ int network_autoconnect(struct network *network, struct scan_bss *bss)
 	if (!network_settings_load(network))
 		return -ENOKEY;
 
-	if (is_rsn) {
-		struct ie_rsn_info rsn;
-
-		memset(&rsn, 0, sizeof(rsn));
-		scan_bss_get_rsn_info(bss, &rsn);
-
-		if (!wiphy_select_cipher(wiphy, rsn.pairwise_ciphers) ||
-				!wiphy_select_cipher(wiphy, rsn.group_cipher)) {
-			l_debug("Cipher mis-match");
-			ret = -ENETUNREACH;
-			goto close_settings;
-		}
-
-		if (rsn.akm_suites & IE_RSN_AKM_SUITE_SAE_SHA256 &&
-				!l_settings_has_key(network->settings,
-						"Security", "Passphrase")) {
-			ret = -ENOKEY;
-			goto close_settings;
-		}
-	}
-
 	/* If no entry, default to Autoconnectable=True */
 	if (!l_settings_get_bool(network->settings, "Settings",
 					"Autoconnect", &is_autoconnectable))
@@ -517,11 +557,24 @@ int network_autoconnect(struct network *network, struct scan_bss *bss)
 	if (!is_autoconnectable)
 		goto close_settings;
 
-	if (network_get_security(network) == SECURITY_PSK) {
-		ret = network_load_psk(network);
+	if (!is_rsn)
+		goto done;
+
+	memset(&rsn, 0, sizeof(rsn));
+	scan_bss_get_rsn_info(bss, &rsn);
+
+	if (!wiphy_select_cipher(wiphy, rsn.pairwise_ciphers) ||
+			!wiphy_select_cipher(wiphy, rsn.group_cipher)) {
+		l_debug("Cipher mis-match");
+		ret = -ENETUNREACH;
+		goto close_settings;
+	}
+
+	if (security == SECURITY_PSK) {
+		ret = network_load_psk(network, __bss_is_sae(bss, &rsn));
 		if (ret < 0)
 			goto close_settings;
-	} else if (network_get_security(network) == SECURITY_8021X) {
+	} else if (security == SECURITY_8021X) {
 		struct l_queue *missing_secrets = NULL;
 
 		ret = eap_check_settings(network->settings, network->secrets,
@@ -539,7 +592,8 @@ int network_autoconnect(struct network *network, struct scan_bss *bss)
 			goto close_settings;
 	}
 
-	return __device_connect_network(network->device, network, bss);
+done:
+	return __station_connect_network(station, network, bss);
 
 close_settings:
 	network_settings_close(network);
@@ -554,7 +608,7 @@ void network_connect_failed(struct network *network)
 	 */
 	if (network_get_security(network) == SECURITY_PSK) {
 		network->update_psk = false;
-		network->ask_psk = true;
+		network->ask_passphrase = true;
 	}
 
 	l_queue_destroy(network->secrets, eap_secret_info_free);
@@ -598,7 +652,7 @@ struct scan_bss *network_bss_find_by_addr(struct network *network,
 struct scan_bss *network_bss_select(struct network *network)
 {
 	struct l_queue *bss_list = network->bss_list;
-	struct wiphy *wiphy = device_get_wiphy(network->device);
+	struct wiphy *wiphy = station_get_wiphy(network->station);
 	const struct l_queue_entry *bss_entry;
 
 	switch (network_get_security(network)) {
@@ -633,7 +687,9 @@ static void passphrase_callback(enum agent_result result,
 				void *user_data)
 {
 	struct network *network = user_data;
+	struct station *station = network->station;
 	struct scan_bss *bss;
+	int r;
 
 	l_debug("result %d", result);
 
@@ -661,20 +717,32 @@ static void passphrase_callback(enum agent_result result,
 
 	l_free(network->psk);
 	network->psk = l_malloc(32);
-	l_free(network->passphrase);
-	network->passphrase = l_strdup(passphrase);
-
-	if (crypto_psk_from_passphrase(passphrase,
+	r = crypto_psk_from_passphrase(passphrase,
 					(uint8_t *) network->info->ssid,
 					strlen(network->info->ssid),
-					network->psk) < 0) {
-		l_error("PMK generation failed.  "
-			"Ensure Crypto Engine is properly configured");
-		dbus_pending_reply(&message, dbus_error_failed(message));
+					network->psk);
+	if (r) {
+		struct l_dbus_message *error;
 
+		l_free(network->psk);
+		network->psk = NULL;
+
+		if (r == -ERANGE || r == -EINVAL)
+			error = dbus_error_invalid_format(message);
+		else {
+			l_error("PSK generation failed: %s.  "
+				"Ensure Crypto Engine is properly configured",
+				strerror(-r));
+			error = dbus_error_failed(message);
+		}
+
+		dbus_pending_reply(&message, error);
 		goto err;
 	}
 
+	l_free(network->passphrase);
+	network->passphrase = l_strdup(passphrase);
+
 	/*
 	 * We need to store the PSK in our permanent store.  However, before
 	 * we do that, make sure the PSK works.  We write to the store only
@@ -682,7 +750,7 @@ static void passphrase_callback(enum agent_result result,
 	 */
 	network->update_psk = true;
 
-	device_connect_network(network->device, network, bss, message);
+	station_connect_network(station, network, bss, message);
 	l_dbus_message_unref(message);
 	return;
 
@@ -690,43 +758,31 @@ err:
 	network_settings_close(network);
 }
 
-static bool bss_is_sae(struct scan_bss *bss)
-{
-	struct ie_rsn_info rsn;
-
-	memset(&rsn, 0, sizeof(rsn));
-	scan_bss_get_rsn_info(bss, &rsn);
-
-	if (rsn.akm_suites & IE_RSN_AKM_SUITE_SAE_SHA256)
-		return true;
-
-	return false;
-}
-
 static struct l_dbus_message *network_connect_psk(struct network *network,
 					struct scan_bss *bss,
 					struct l_dbus_message *message)
 {
-	struct device *device = network->device;
-
-	l_debug("");
-
-	if (network_settings_load(network))
-		network_load_psk(network);
-	else
-		network->settings = l_settings_new();
-
-	l_debug("ask_psk: %s", network->ask_psk ? "true" : "false");
-
+	struct station *station = network->station;
 	/*
 	 * A legacy psk file may only contain the PreSharedKey entry. For SAE
 	 * networks the raw Passphrase is required. So in this case where
-	 * the psk is found but no passphrase, we ask the agent. In this case
-	 * the psk file will be re-written to contain the raw passphrase.
+	 * the psk is found but no Passphrase, we ask the agent.  The psk file
+	 * will then be re-written to contain the raw passphrase.
 	 */
-	if (network->ask_psk || !network->psk ||
-			(!network->passphrase && bss_is_sae(bss))) {
-		network->ask_psk = false;
+	bool need_passphrase = bss_is_sae(bss);
+
+	l_debug("ask_passphrase: %s",
+		network->ask_passphrase ? "true" : "false");
+
+	if  (!network_settings_load(network)) {
+		network->settings = l_settings_new();
+		network->ask_passphrase = true;
+	} else if (!network->ask_passphrase)
+		network->ask_passphrase =
+			network_load_psk(network, need_passphrase) < 0;
+
+	if (network->ask_passphrase) {
+		network->ask_passphrase = false;
 
 		network->agent_request =
 			agent_request_passphrase(network->object_path,
@@ -736,7 +792,7 @@ static struct l_dbus_message *network_connect_psk(struct network *network,
 		if (!network->agent_request)
 			return dbus_error_no_agent(message);
 	} else
-		device_connect_network(device, network, bss, message);
+		station_connect_network(station, network, bss, message);
 
 	return NULL;
 }
@@ -935,6 +991,7 @@ static struct l_dbus_message *network_connect_8021x(struct network *network,
 						struct scan_bss *bss,
 						struct l_dbus_message *message)
 {
+	struct station *station = network->station;
 	int r;
 	struct l_queue *missing_secrets = NULL;
 	struct l_dbus_message *reply;
@@ -967,7 +1024,7 @@ static struct l_dbus_message *network_connect_8021x(struct network *network,
 			goto error;
 		}
 
-		device_connect_network(network->device, network, bss, message);
+		station_connect_network(station, network, bss, message);
 
 		return NULL;
 	}
@@ -992,12 +1049,12 @@ static struct l_dbus_message *network_connect(struct l_dbus *dbus,
 						void *user_data)
 {
 	struct network *network = user_data;
-	struct device *device = network->device;
+	struct station *station = network->station;
 	struct scan_bss *bss;
 
 	l_debug("");
 
-	if (device_is_busy(device))
+	if (station_is_busy(station))
 		return dbus_error_busy(message);
 
 	/*
@@ -1015,7 +1072,7 @@ static struct l_dbus_message *network_connect(struct l_dbus *dbus,
 	case SECURITY_PSK:
 		return network_connect_psk(network, bss, message);
 	case SECURITY_NONE:
-		device_connect_network(device, network, bss, message);
+		station_connect_network(station, network, bss, message);
 		return NULL;
 	case SECURITY_8021X:
 		if (!network_settings_load(network))
@@ -1030,7 +1087,7 @@ static struct l_dbus_message *network_connect(struct l_dbus *dbus,
 void network_connect_new_hidden_network(struct network *network,
 						struct l_dbus_message *message)
 {
-	struct device *device = network->device;
+	struct station *station = network->station;
 	struct scan_bss *bss;
 	struct l_dbus_message *error;
 
@@ -1058,7 +1115,7 @@ void network_connect_new_hidden_network(struct network *network,
 		error = network_connect_psk(network, bss, message);
 		break;
 	case SECURITY_NONE:
-		device_connect_network(device, network, bss, message);
+		station_connect_network(station, network, bss, message);
 		return;
 	default:
 		error = dbus_error_not_supported(message);
@@ -1091,9 +1148,10 @@ static bool network_property_is_connected(struct l_dbus *dbus,
 					void *user_data)
 {
 	struct network *network = user_data;
+	struct station *station = network->station;
 	bool connected;
 
-	connected = device_get_connected_network(network->device) == network;
+	connected = station_get_connected_network(station) == network;
 	l_dbus_message_builder_append_basic(builder, 'b', &connected);
 	return true;
 }
@@ -1104,9 +1162,11 @@ static bool network_property_get_device(struct l_dbus *dbus,
 					void *user_data)
 {
 	struct network *network = user_data;
+	struct station *station = network->station;
+	struct netdev *netdev = station_get_netdev(station);
 
 	l_dbus_message_builder_append_basic(builder, 'o',
-					device_get_path(network->device));
+						netdev_get_path(netdev));
 
 	return true;
 }
@@ -1179,7 +1239,7 @@ bool network_register(struct network *network, const char *path)
 		l_info("Unable to register %s interface",
 						L_DBUS_INTERFACE_PROPERTIES);
 
-	network->object_path = strdup(path);
+	network->object_path = l_strdup(path);
 
 	return true;
 }
@@ -1237,18 +1297,14 @@ int network_rank_compare(const void *a, const void *b, void *user)
 	return network->rank - new_network->rank;
 }
 
-void network_rank_update(struct network *network)
+void network_rank_update(struct network *network, bool connected)
 {
-	bool connected;
-	int rank;
-
 	/*
 	 * Theoretically there may be difference between the BSS selection
 	 * here and in network_bss_select but those should be rare cases.
 	 */
 	struct scan_bss *best_bss = l_queue_peek_head(network->bss_list);
-
-	connected = device_get_connected_network(network->device) == network;
+	int rank;
 
 	/*
 	 * The rank should separate networks into four groups that use
@@ -1279,12 +1335,12 @@ void network_rank_update(struct network *network)
 	network->rank = rank;
 }
 
-static void emit_known_network_changed(struct device *device, void *user_data)
+static void emit_known_network_changed(struct station *station, void *user_data)
 {
 	struct network_info *info = user_data;
 	struct network *network;
 
-	network = device_network_find(device, info->ssid, info->type);
+	network = station_network_find(station, info->ssid, info->type);
 	if (!network)
 		return;
 
@@ -1307,7 +1363,7 @@ struct network_info *network_info_add_known(const char *ssid,
 	if (network) {
 		/* Promote network to is_known */
 		network->is_known = true;
-		__iwd_device_foreach(emit_known_network_changed, network);
+		station_foreach(emit_known_network_changed, network);
 		return network;
 	}
 
@@ -1319,23 +1375,23 @@ struct network_info *network_info_add_known(const char *ssid,
 	return network;
 }
 
-static void disconnect_no_longer_known(struct device *device, void *user_data)
+static void disconnect_no_longer_known(struct station *station, void *user_data)
 {
 	struct network_info *info = user_data;
 	struct network *network;
 
-	network = device_get_connected_network(device);
+	network = station_get_connected_network(station);
 
 	if (network && network->info == info)
-		device_disconnect(device);
+		station_disconnect(station);
 }
 
 void network_info_forget_known(struct network_info *network)
 {
 	network->is_known = false;
 
-	__iwd_device_foreach(emit_known_network_changed, network);
-	__iwd_device_foreach(disconnect_no_longer_known, network);
+	station_foreach(emit_known_network_changed, network);
+	station_foreach(disconnect_no_longer_known, network);
 
 	/*
 	 * Network is no longer a Known Network, see if we still need to
